# -*- coding: utf-8 -*-
##############################################################################
#
# Yoshinori OKUJI <yo@nexedi.com>
#
# Copyright (C) 2003 Nexedi SARL
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. ?See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA ?02111-1307, USA.
#
##############################################################################

from lxml import etree

import sys
import getopt
import os
from StringIO import StringIO
import re
import codecs
from copy import deepcopy
from interfaces.erp5diff import IERP5Diff
import zope.interface

class ERP5Diff:
  """
    Make a difference between two XML documents using XUpdate.
    Use some assumptions in ERP5's data representation.

    The strategy is:
      1. Find a matching element among elements of the other XML document at the same depth.
      2. Use the first matching element, even if there can be other better elements.
      3. Assume that two elements are matching, if the tag names are identical. If either of
         them has an attribute 'id', the values of the attributes 'id' also must be identical.
      4. Don't use xupdate:rename for elements. It should be quite rare to rename tag names
         in ERP5, and it is too complicated to support this renaming.
      5. Ignore some types of nodes, such as EntityReference and Comment, because they are not
         used in ERP5 XML documents.
  """

  # Declarative interfaces
  zope.interface.implements(IERP5Diff,)

  __version__ = 0.2

  def __init__(self):
    """
      Initialize itself.
    """
    self._verbose = 0
    self._result = None
    self._ns = 'http://www.xmldb.org/xupdate'

  def setVerbosity(self, verbose):
    """
      Set the verbosity.
    """
    self._verbose = verbose

  def _p(self, msg):
    """
      Print a message only if being verbose.
    """
    if self._verbose:
      sys.stderr.write(str(msg) + os.linesep)

  def _makeDocList(self, *args):
    """
      Make a list of Document objects.
    """
    doc_list = []
    for a in args:
      if type(a) == type(''):
        doc_list.append(etree.XML(a))
      else:
        element_tree = etree.parse(a)
        doc_list.append(element_tree.getroot())
    return doc_list

  def _concatPath(self, p1, p2, separator='/'):
    """
      Concatenate 'p1' and 'p2'. Add a separator between them,
      only if 'p1' does not end with a separator.
    """
    if p1.endswith(separator):
      return p1 + p2
    return p1 + separator + p2

  def _getResultRoot(self):
    """
      Return the root element of the result document.
    """
    return self._result
    #return self._result.getroottree()

  def _hasChildren(self, element):
    """
      Check whether the element has any children
    """
    return bool(len(element))

  def _xupdateAppendAttributes(self, dict, path):
    """
      Append attrib to the element at 'path'.
    """
    root = self._getResultRoot()
    append_element = etree.Element('{%s}append' % self._ns, nsmap=root.nsmap)
    append_element.attrib['select'] = path
    for name, val in dict.iteritems():
      attr_element = etree.Element('{%s}attribute' % self._ns, nsmap=root.nsmap)
      attr_element.attrib['name'] = name
      attr_element.text = val
      append_element.append(attr_element)
    root.append(append_element)

  def _xupdateRemoveAttribute(self, name, path):
    """
      Remove an attribute from the element at 'path'.
    """
    root = self._getResultRoot()
    remove_element = etree.Element('{%s}remove' % self._ns, nsmap=root.nsmap)
    remove_element.attrib['select'] = self._concatPath(path, 'attribute::' + name)
    root.append(remove_element)

  def _xupdateUpdateAttribute(self, name, val, path):
    """
      Update the value of an attribute of the element at 'path'.
    """
    root = self._getResultRoot()
    update_element = etree.Element('{%s}update' % self._ns, nsmap=root.nsmap)
    update_element.attrib['select'] = self._concatPath(path, 'attribute::' + name)
    update_element.text = val
    root.append(update_element)

  def _xupdateRenameElement(self, name, path):
    """
      Rename an existing element at 'path'.
    """
    root = self._getResultRoot()
    rename_element = etree.Element('{%s}rename' % self._ns, nsmap=root.nsmap)
    rename_element.attrib['select'] = path
    rename_element.text = name
    root.append(rename_element)

  def _xupdateUpdateElement(self, element, path):
    """
      Update the contents of an element at 'path' to that of 'element'.
    """
    root = self._getResultRoot()
    update_element = etree.Element('{%s}update' % self._ns, nsmap=root.nsmap)
    update_element.attrib['select'] = path 
    if self._hasChildren(element):
      for child in element:
        clone_node = deepcopy(child)
        update_element.append(clone_node)
    else:
      update_element.text = element.text
    root.append(update_element)

  def _xupdateRemoveElement(self, path):
    """
      Remove an element at 'path'.
    """
    root = self._getResultRoot()
    remove_element = etree.Element('{%s}remove' % self._ns, nsmap=root.nsmap)
    remove_element.attrib['select'] = path
    root.append(remove_element)

  def _xupdateInsertBefore(self, element_list, path):
    """
      Insert elements before the element at 'path'.
    """
    root = self._getResultRoot()
    insert_element = etree.Element('{%s}insert-before' % self._ns, nsmap=root.nsmap)
    insert_element.attrib['select'] = path
    for element in element_list:
      child_element = etree.Element('{%s}element' % self._ns, nsmap=root.nsmap)
      child_element.attrib['name'] = element.tag
      attr_map = element.attrib
      for name, value in attr_map.items():
        attr_element = etree.Element('{%s}attribute' % self._ns, nsmap=root.nsmap)
        attr_element.attrib['name'] = name
        attr_element.text = value
        child_element.append(attr_element)
      for child in element:
        clone_node = deepcopy(child)
        child_element.append(clone_node)
      if self._hasChildren(child_element):
        child_element[-1].tail = element.text
      insert_element.append(child_element)
    root.append(insert_element)

  def _xupdateAppendElements(self, element_list, path):
    """
      Append elements to the element at 'path'.
    """
    root = self._getResultRoot()
    append_element = etree.Element('{%s}append' % self._ns, nsmap=root.nsmap)
    append_element.attrib['select'] = path
    for element in element_list:
      child_element = etree.Element('{%s}element' % self._ns, nsmap=root.nsmap)
      child_element.attrib['name'] = element.tag
      attr_map = element.attrib
      for name, value in attr_map.items():
        attr_element = etree.Element('{%s}attribute' % self._ns, nsmap=root.nsmap)
        attr_element.attrib['name'] = name
        attr_element.text = value
        child_element.append(attr_element)
      for child in element:
        clone_node = deepcopy(child)
        child_element.append(clone_node)
      if self._hasChildren(child_element):
        child_element[-1].tail = element.text
      append_element.append(child_element)
    root.append(append_element)

  def _testElements(self, element1, element2):
    """
      Test if two given elements are matching. Matching does not mean that they are identical.
    """
    # Make sure that they are elements.
    if type(element1) !=  type(element2) or type(element1) != etree._Element:
      return False

    if element1.tag != element2.tag:
      return False

    id_list = []
    for attr_map in (element1.attrib, element2.attrib):
      for name, value in attr_map.items():
        if name == 'id':
          id_list.append(value)
          break

    if len(id_list) == 0:
      return True
    if len(id_list) == 1:
      return False
    return (id_list[0] == id_list[1])

  def _testAttributes(self, element1, element2, path):
    """
      Test attrib of two given elements. Add differences, if any.
    """
    # Make a list of dictionaries of the attributes.
    dict_list = []
    for attr_map in (element1.attrib, element2.attrib):
      d = {}
      for name, value in attr_map.items():
        d[name] = value
      dict_list.append(d)
    dict1, dict2 = dict_list

    # Find all added or removed or changed attrib.
    for name1, val1 in dict1.iteritems():
      if name1 in dict2:
        if val1 != dict2[name1]:
          # The value is different.
          self._xupdateUpdateAttribute(name1, dict2[name1], path)
        # Mark this attribute.
        dict2[name1] = None
      else:
        # This attribute is removed.
        self._xupdateRemoveAttribute(name1, path)
    d = {}
    for name2, val2 in dict2.iteritems():
      if val2 is not None:
        # This attribute is added.
        d[name2] = val2
    if d != {}:
      self._xupdateAppendAttributes(d, path)

  def _checkEmptiness(self, element):
    """
      Check if an element has child values.
    """
    for child in element:
      if type(child) == etree._Element:
        return False
    if element.text is not None:
      return False
    return True

  def _checkIgnoreText(self, element):
    """
      Determine if text should be ignored by heuristics,
      because ERP5 does not define any schema at the moment.
    """
    for child in element:
      if type(child) == etree._Element:
        return True
    return False

  def _makeRelativePathList(self, element_list):
    """
      Make a list of relative paths from a list of elements.
    """
    num_map = {}
    count_map = {}
    for element in element_list:
      if element.tag in num_map:
        num_map[element.tag] += 1
      else:
        num_map[element.tag] = 1
        count_map[element.tag] = 0

    path_list = []
    for element in element_list:
      # Check if this element has an attribute 'id'.s
      id_val = None
      attr_map = element.attrib
      for name, value in attr_map.items():
        if name == 'id':
          id_val = value
          break

      if id_val is not None:
        # If an attribute 'id' is present, uses the attribute for convenience.
        path_list.append("%s[@id='%s']" % (element.tag, id_val))
        # Increase the count, for a case where other elements with the same tag name do not have
        # 'id' attrib.
        count_map[element.tag] += 1
      elif num_map[element.tag] > 1:
        path_list.append('%s[%d]' % (element.tag, count_map[element.tag]))
        count_map[element.tag] += 1
      else:
        path_list.append(element.tag)

    return path_list

  def _aggregateElements(self, element):
    """
      Aggregate child elements of an element into a list.
    """
    element_list = []
    for child in element:
      if type(child) == etree._Element:
        element_list.append(child)
    return element_list

  def _aggregateText(self, element):
    """
      Aggregate child text nodes of an element into a single string.
    """
    text = ''
    if not self._hasChildren(element):
      return element.text
    for child in element:
      if type(child) == etree._Element:
        text += child.text
    return text

  def _compareChildNodes(self, old_element, new_element, path):
    """
      Compare children of two elements, and add differences into the result, if any.
      Call itself recursively, if these elements have grandchilden.
    """
    self._p("Comparing %s with %s at %s..." % (repr(old_element), repr(new_element), path))

    # First, determine if they are empty.
    old_is_empty = self._checkEmptiness(old_element)
    new_is_empty = self._checkEmptiness(new_element)

    if old_is_empty and new_is_empty:
      # Nothing to do.
      self._p("Both are empty.")
      pass
    elif old_is_empty or new_is_empty:
      # Perhaps they are very different.
      self._p("One of them is empty, so just update all the contents.")
      self._xupdateUpdateElement(new_element, path)
    else:
      # Second, determine if text should be ignored.
      old_ignore_text = self._checkIgnoreText(old_element)
      new_ignore_text = self._checkIgnoreText(new_element)

      if old_ignore_text != new_ignore_text:
        # This means that the semantics of this element is quite different.
        self._p("One of them has only text and the other does not, so just update all the contents.")
        self._xupdateUpdateElement(new_element, path)
      elif not old_ignore_text:
        # The contents are only text.
        self._p("Both have only text.")
        old_text = self._aggregateText(old_element)
        new_text = self._aggregateText(new_element)
        if old_text != new_text:
          self._p("They differ, so update the elements.")
          self._xupdateUpdateElement(new_element, path)
      else:
        # The contents are elements.
        self._p("Both have elements.")
        old_list = self._aggregateElements(old_element)
        path_list = self._makeRelativePathList(old_list)
        new_list = self._aggregateElements(new_element)
        new_start = 0
        new_len = len(new_list)
        for old_node, node_path in zip(old_list, path_list):
          child_path = self._concatPath(path, node_path)
          for new_current in range(new_start, new_len):
            new_node = new_list[new_current]
            if self._testElements(old_node, new_node):
              self._testAttributes(old_node, new_node, child_path)
              self._compareChildNodes(old_node, new_node, child_path)
              if new_current > new_start:
                # There are skipped nodes in the new children.
                self._xupdateInsertBefore(new_list[new_start:new_current], child_path)
              new_start = new_current + 1
              break
          else:
            # There is no matching node. So this element must be removed.
            self._xupdateRemoveElement(child_path)
        if new_len > new_start:
          # There are remaining nodes in the new children.
          self._xupdateAppendElements(new_list[new_start:new_len], path)

  def compare(self, old_xml, new_xml):
    """
      Compare two given XML documents.
      If an argument is a string, it is assumed to be a XML document itself.
      Otherwise, it is assumed to be a file object which contains a XML document.
    """
    old_doc, new_doc = self._makeDocList(old_xml, new_xml)
    old_root_element = old_doc #.getroottree() #old_doc.documentElement
    new_root_element = new_doc #.getroottree() #new_doc.documentElement
    try:
      if self._result is not None:
        self._result = None 
      self._result = etree.Element('{%s}modifications' % self._ns, nsmap={'xupdate':self._ns})
      self._result.set('version', '1.0')
      if self._testElements(old_root_element, new_root_element):
        self._testAttributes(old_root_element, new_root_element, '/')
        self._compareChildNodes(old_root_element, new_root_element, '/')
      else:
        # These XML documents seem to be completely different...
        if old_root_element.tag != new_root_element.tag:
          self._xupdateRenameElement(new_root_element.tag, '/')
        self._testAttributes(old_root_element, new_root_element, '/')
        self._xupdateUpdateElement(new_root_element, '/')
    finally:
      del old_doc
      del new_doc

  def output(self, output_file=None):
    """
      Output the result of parsing XML documents to 'output_file'.
      If it is not specified, stdout is assumed.
    """
    if output_file is None:
      output_file = sys.stdout
    # stream
    xml = etree.tostring(self._result, encoding='utf-8', pretty_print=True)
    output_file.write(xml)

  def outputString(self):
    """
      Return the result as a string object.
    """
    io = StringIO()
    self.output(io)
    ret = io.getvalue()
    io.close()
    return ret

def main():
  """
    The main routine of ERP5Diff.
  """
  try:
    opts, args = getopt.getopt(sys.argv[1:], "ho:v", ["help", "output=", "verbose"])
  except getopt.GetoptError, msg:
    print msg
    print "Try ``erp5diff --help'' for more information."
    sys.exit(2)
  output = None
  verbose = 0
  for o, a in opts:
    if o == "-v":
      verbose = 1
    elif o in ("-h", "--help"):
      print '''Usage: erp5diff [OPTION]... OLD_XML NEW_XML
Make a difference between two XML documents in XUpdate format.

    -h, --help          display this message and exit
    -o, --output=FILE   output the result to the file FILE
    -v, --verbose       print verbose messages

Report bugs to <yo@nexedi.com>.'''
      sys.exit()
    elif o in ("-o", "--output"):
      output = a

  if len(args) != 2:
    if len(args) > 2:
      print "Too many arguments."
    else:
      print "Too few arguments."
    print "Try ``erp5diff --help'' for more information."
    sys.exit(2)

  d = ERP5Diff()
  d.setVerbosity(verbose)

  old_xml = open(args[0])
  new_xml = open(args[1])
  d.compare(old_xml, new_xml)
  old_xml.close()
  new_xml.close()

  try:
    if output is not None:
      file = open(output, 'w')
    else:
      file = None
    d.output(file)
  except:
    if output is not None:
      file.close()
      os.remove(output)
    raise
  else:
    if file is not None:
      file.close()

  sys.exit()

if __name__ == '__main__':
  main()
